@@ -14,7 +14,8 @@
 #define _LIBNVME_API_TYPES_H
 
 #include <stdbool.h>
-#include "types.h"
+
+#include <nvme/types.h>
 
 /*
  * _args struct definitions. These are used by both the ioctl-based and
@@ -511,6 +512,7 @@ struct nvme_get_property_args {
  * @owpass:	Overwrite pass count
  * @oipbp:	Set to overwrite invert pattern between passes
  * @nodas:	Set to not deallocate blocks after sanitizing
+ * @emvs:	Set to enter media verification state
  */
 struct nvme_sanitize_nvm_args {
 	__u32 *result;
@@ -523,6 +525,7 @@ struct nvme_sanitize_nvm_args {
 	__u8 owpass;
 	bool oipbp;
 	bool nodas;
+	bool emvs;
 };
 
 /**
@@ -960,4 +963,141 @@ struct nvme_dim_args {
 	__u8	tas;
 };
 
+/**
+ * struct nvme_lm_cdq_args - Arguments for Controller Data Queue (CDQ) command
+ * @result:	Set on completion to the command's CQE DWORD 0 controller response
+ * @data:	Pointer to data
+ * @args_size:	Length of structure
+ * @fd:		File descriptor of nvme device
+ * @timeout:	Timeout in ms
+ * @mos:	Management Operation Specific (MOS): This field is specific to the SEL type
+ * @cntlid:	Controller ID: For Create CDQ, specifies the target migratable controller
+ * @cdqid:	Controller Data Queue ID (CDQID): For Create CDQ, this field is the CDQID created
+ *		by the controller if no error is present. For Delete CDQ, this field is the CDQID
+ *		to delete.
+ * @sel:	Select (SEL): This field specifies the type of management operation to perform.
+ * @sz:		For Create CDQ, specifies the size of CDQ, in dwords
+ */
+struct nvme_lm_cdq_args {
+	__u32	*result;
+	void	*data;
+	int	args_size;
+	int	fd;
+	__u32	timeout;
+	__u16	mos;
+	__u16	cntlid;
+	__u16	cdqid;
+	__u8	sel;
+	__u8	sz;
+};
+
+/**
+ * struct nvme_lm_track_send_args - Arguments for the Track Send command
+ * @result:	Set on completion to the command's CQE DWORD 0 controller response
+ * @args_size:	Length of structure
+ * @fd:		File descriptor of nvme device
+ * @timeout:	Timeout in ms
+ * @mos:	Management Operation Specific (MOS): This field is specific to the SEL type
+ * @cdqid:	Controller Data Queue ID (CDQID)
+ * @sel:	Select (SEL): This field specifies the type of management operation to perform
+ */
+struct nvme_lm_track_send_args {
+	__u32	*result;
+	int	args_size;
+	int	fd;
+	__u32	timeout;
+	__u16	mos;
+	__u16	cdqid;
+	__u8	sel;
+};
+
+/**
+ * struct nvme_lm_migration_send_args - Arguments for the Migration Send command
+ * @offset:	Offset: This field specifies the offset, in bytes, within the data available to be
+ *		returned and specifies the starting point for that data for what is actually
+ *		returned to the host.
+ * @result:	Set on completion to the command's CQE DWORD 0 controller response
+ * @data:	Pointer to data
+ * @args_size:	Length of structure
+ * @fd:		File descriptor of nvme device
+ * @timeout:	Timeout in ms
+ * @numd:	Number of Dwords (NUMD): This field specifies the number of dwords being transferred
+ * @mos:	Management Operation Specific (MOS): This field is specific to the SEL type
+ * @cntlid:	Controller ID: This field specifies the identifier of the controller to which the
+ *		operation is performed.
+ * @csuuidi:	Controller State UUID Index (CSUUIDI): A non-zero value in this field specifies the
+ *		index to a specific entry in the Vendor Specific Controller State UUID Supported.
+ *		list of the Supported Controller State Formats data structure.
+ * @sel:	Select (SEL): This field specifies the type of management operation to perform.
+ * @uidx:	UUID Index (UIDX): If this field is set to a non-zero value, then the value of this
+ *		field is the index of a UUID in the UUID List (refer to Figure 320) that is used by
+ *		the command.
+ * @stype:	Suspend Type (STYPE): This field specifies the type of suspend.
+ * @seqind:	Sequence Identifier (SEQIND): This field identified the sequences of this Migration
+ *		Send command in relation to other Migration Send commands.
+ * @csvi:	Controller State Version Index (CSVI): A non-zero value in this field specifies the
+ *		index to a specific entry in the NVMe Controller State Version list of the Supported
+ *		Controller State Formats data structure.
+ * @dudmq:	Delete User Data Migration Queue (DUDMQ): If set, the migration queue is deleted
+ *		is deleted as part of the Suspend operation. If cleared, it is retained.
+ */
+struct nvme_lm_migration_send_args {
+	__u64	offset;
+	__u32	*result;
+	void	*data;
+	int	args_size;
+	int	fd;
+	__u32	timeout;
+	__u32	numd;
+	__u16	mos;
+	__u16	cntlid;
+	__u16	csuuidi;
+	__u8	sel;
+	__u8	uidx;
+	__u8	stype;
+	__u8	seqind;
+	__u8	csvi;
+	bool	dudmq;
+};
+
+/**
+ * struct nvme_lm_migration_recv_args - Arguments for the Migration Receive command
+ * @offset:	Offset: This field specifies the offset, in bytes, within the data available to be
+ *		returned and specifies the starting point for that data for what is actually
+ *		returned to the host.
+ * @result:	Set on completion to the command's CQE DWORD 0 controller response
+ * @data:	Pointer to data
+ * @args_size:	Length of structure
+ * @fd:		File descriptor of nvme device
+ * @timeout:	Timeout in ms
+ * @numd:	Number of Dwords (NUMD): This field specifies the number of dwords to return. This
+ *		is a 0's based value.
+ * @mos:	Management Operation Specific (MOS): This field is specific to the SEL type
+ * @cntlid:	Controller ID: This field specifies the identifier of the controller to which the
+ *		operation is performed.
+ * @csuuidi:	Controller State UUID Index (CSUUIDI): A non-zero value in this field specifies the
+ *		index to a specific entry in the Vendor Specific Controller State UUID Supported.
+ *		list of the Supported Controller State Formats data structure.
+ * @sel:	Select (SEL): This field specifies the type of management operation to perform
+ * @uidx:	UUID Index (UIDX): If this field is set to a non-zero value, then the value of this
+ *		field is the index of a UUID in the UUID List (refer to Figure 320) that is used by
+ *		the command.
+ * @csuidxp:	Controller State UUID Index Parameter (CSUIDXP): This field is vendor specific.
+ */
+struct nvme_lm_migration_recv_args {
+	__u64	offset;
+	__u32	*result;
+	void	*data;
+	int	args_size;
+	int	fd;
+	__u32	timeout;
+	__u32	numd;
+	__u16	mos;
+	__u16	cntlid;
+	__u16	csuuidi;
+	__u8	sel;
+	__u8	uidx;
+	__u8	csuidxp;
+};
+
 #endif /* _LIBNVME_API_TYPES_H */
