@@ -14,6 +14,7 @@
 #include <unistd.h>
 
 #include <ccan/array_size/array_size.h>
+#include <ccan/ccan/minmax/minmax.h>
 #include <ccan/endian/endian.h>
 
 #include "log.h"
@@ -529,6 +530,31 @@ static void nvme_mi_admin_init_resp(struct nvme_mi_resp *resp,
 	resp->hdr_len = sizeof(*hdr);
 }
 
+static void nvme_mi_control_init_req(struct nvme_mi_req *req,
+				    struct nvme_mi_control_req *control_req,
+				    __u8 opcode, __u16 cpsp)
+{
+	memset(req, 0, sizeof(*req));
+	memset(control_req, 0, sizeof(*control_req));
+
+	control_req->hdr.type = NVME_MI_MSGTYPE_NVME;
+	control_req->hdr.nmp = (NVME_MI_ROR_REQ << 7) |
+		(NVME_MI_MT_CONTROL << 3); /* we always use command slot 0 */
+	control_req->opcode = opcode;
+	control_req->cpsp = cpu_to_le16(cpsp);
+
+	req->hdr = &control_req->hdr;
+	req->hdr_len = sizeof(*control_req);
+}
+
+static void nvme_mi_control_init_resp(struct nvme_mi_resp *resp,
+				    struct nvme_mi_control_resp *control_resp)
+{
+	memset(resp, 0, sizeof(*resp));
+	resp->hdr = &control_resp->hdr;
+	resp->hdr_len = sizeof(*control_resp);
+}
+
 static int nvme_mi_admin_parse_status(struct nvme_mi_resp *resp, __u32 *result)
 {
 	struct nvme_mi_admin_resp_hdr *admin_hdr;
@@ -580,6 +606,26 @@ static int nvme_mi_admin_parse_status(struct nvme_mi_resp *resp, __u32 *result)
 	return nvme_status;
 }
 
+static int nvme_mi_control_parse_status(struct nvme_mi_resp *resp, __u16 *cpsr)
+{
+	struct nvme_mi_control_resp *control_resp;
+
+	if (resp->hdr_len < sizeof(*control_resp)) {
+		errno = -EPROTO;
+		return -1;
+	}
+	control_resp = (struct nvme_mi_control_resp *)resp->hdr;
+
+	if (control_resp->status)
+		return control_resp->status |
+			(NVME_STATUS_TYPE_MI << NVME_STATUS_TYPE_SHIFT);
+
+	if (cpsr)
+		*cpsr = le16_to_cpu(control_resp->cpsr);
+
+	return control_resp->status;
+}
+
 int nvme_mi_admin_xfer(nvme_mi_ctrl_t ctrl,
 		       struct nvme_mi_admin_req_hdr *admin_req,
 		       size_t req_data_size,
@@ -589,6 +635,7 @@ int nvme_mi_admin_xfer(nvme_mi_ctrl_t ctrl,
 {
 	struct nvme_mi_resp resp;
 	struct nvme_mi_req req;
+	__u32 dlen, doff;
 	int rc;
 
 	/* length/offset checks. The common _submit() API will do further
@@ -647,8 +694,17 @@ int nvme_mi_admin_xfer(nvme_mi_ctrl_t ctrl,
 
 	/* limit the response size, specify offset */
 	admin_req->flags = 0x3;
-	admin_req->dlen = cpu_to_le32(resp.data_len & 0xffffffff);
-	admin_req->doff = cpu_to_le32(resp_data_offset & 0xffffffff);
+
+	/* dlen and doff have different interpretations depending on the data direction */
+	if (req_data_size) {
+		dlen = req_data_size & 0xffffffff;
+		doff = 0;
+	} else {
+		dlen = *resp_data_size & 0xffffffff;
+		doff = resp_data_offset & 0xffffffff;
+	}
+	admin_req->dlen = cpu_to_le32(dlen);
+	admin_req->doff = cpu_to_le32(doff);
 
 	rc = nvme_mi_submit(ctrl->ep, &req, &resp);
 	if (rc)
@@ -812,6 +868,29 @@ int nvme_mi_admin_identify_partial(nvme_mi_ctrl_t ctrl,
 	return 0;
 }
 
+int nvme_mi_control(nvme_mi_ep_t ep, __u8 opcode,
+		    __u16 cpsp, __u16 *result_cpsr)
+{
+	struct nvme_mi_control_resp control_resp;
+	struct nvme_mi_control_req control_req;
+	struct nvme_mi_resp resp;
+	struct nvme_mi_req req;
+	int rc = 0;
+
+	nvme_mi_control_init_req(&req, &control_req, opcode, cpsp);
+	nvme_mi_control_init_resp(&resp, &control_resp);
+
+	rc = nvme_mi_submit(ep, &req, &resp);
+	if (rc)
+		return rc;
+
+	rc = nvme_mi_control_parse_status(&resp, result_cpsr);
+	if (rc)
+		return rc;
+
+	return 0;
+}
+
 /* retrieves a MCTP-messsage-sized chunk of log page data. offset and len are
  * specified within the args->data area. The `offset` parameter is a relative
  * offset to the args->lpo !
@@ -943,7 +1022,7 @@ static int read_ana_chunk(nvme_mi_ctrl_t ctrl, enum nvme_log_ana_lsp lsp, bool r
 	}
 
 	while (*read < to_read) {
-		__u32 len = min(log_end - *read, NVME_LOG_PAGE_PDU_SIZE);
+		__u32 len = min_t(__u32, log_end - *read, NVME_LOG_PAGE_PDU_SIZE);
 		int ret;
 
 		ret = nvme_mi_admin_get_log_ana(ctrl, lsp, rae,
@@ -1521,6 +1600,72 @@ static int nvme_mi_read_data(nvme_mi_ep_t ep, __u32 cdw0,
 	return 0;
 }
 
+int nvme_mi_mi_xfer(nvme_mi_ep_t ep,
+		       struct nvme_mi_mi_req_hdr *mi_req,
+		       size_t req_data_size,
+		       struct nvme_mi_mi_resp_hdr *mi_resp,
+		       size_t *resp_data_size)
+{
+	int rc;
+	struct nvme_mi_req req;
+	struct nvme_mi_resp resp;
+
+	/* There is nothing in the spec to define this limit but going with the limits
+	 * from the admin message types for DLEN seems like a reasonable starting point
+	 * to check for coding errors
+	 */
+	const size_t mi_data_xfer_size_limit = 4096;
+
+	/* length/offset checks. The common _submit() API will do further
+	 * checking on the message lengths too, so these are kept specific
+	 * to the requirements of the particular command set
+	 */
+
+	if (*resp_data_size > mi_data_xfer_size_limit) {
+		errno = EINVAL;
+		return -1;
+	}
+
+	/* request and response lengths & offset must be aligned */
+	if ((req_data_size & 0x3) ||
+	    (*resp_data_size & 0x3)) {
+		errno = EINVAL;
+		return -1;
+	}
+
+	/* bidirectional not permitted */
+	if (req_data_size && *resp_data_size) {
+		errno = EINVAL;
+		return -1;
+	}
+
+	mi_req->hdr.type = NVME_MI_MSGTYPE_NVME;
+	mi_req->hdr.nmp = (NVME_MI_ROR_REQ << 7) |
+				(NVME_MI_MT_MI << 3);
+
+	memset(&req, 0, sizeof(req));
+	req.hdr = &mi_req->hdr;
+	req.hdr_len = sizeof(*mi_req);
+	req.data = mi_req + 1;
+	req.data_len = req_data_size;
+
+	nvme_mi_calc_req_mic(&req);
+
+	memset(&resp, 0, sizeof(resp));
+	resp.hdr = &mi_resp->hdr;
+	resp.hdr_len = sizeof(*mi_resp);
+	resp.data = mi_resp + 1;
+	resp.data_len = *resp_data_size;
+
+	rc = nvme_mi_submit(ep, &req, &resp);
+	if (rc)
+		return rc;
+
+	*resp_data_size = resp.data_len;
+
+	return 0;
+}
+
 int nvme_mi_mi_read_mi_data_subsys(nvme_mi_ep_t ep,
 				   struct nvme_mi_read_nvm_ss_info *s)
 {
