@@ -11,7 +11,7 @@
 
 #include <ifaddrs.h>
 
-#include "types.h"
+#include <nvme/types.h>
 
 /**
  * DOC: util.h
@@ -337,6 +337,16 @@ static inline void nvme_feature_decode_write_atomicity(__u32 value, bool *dn)
 #define NVME_FEAT_AE_PLA(v)		NVME_GET(v, FEAT_AE_PLA)
 #define NVME_FEAT_AE_LBAS(v)		NVME_GET(v, FEAT_AE_LBAS)
 #define NVME_FEAT_AE_EGA(v)		NVME_GET(v, FEAT_AE_EGA)
+#define NVME_FEAT_AE_NNSSHDN(v)		NVME_GET(v, FEAT_AE_NNSSHDN)
+#define NVME_FEAT_AE_TTHRY(v)		NVME_GET(v, FEAT_AE_TTHRY)
+#define NVME_FEAT_AE_RASSN(v)		NVME_GET(v, FEAT_AE_RASSN)
+#define NVME_FEAT_AE_RGRP0(v)		NVME_GET(v, FEAT_AE_RGRP0)
+#define NVME_FEAT_AE_ANSAN(v)		NVME_GET(v, FEAT_AE_ANSAN)
+#define NVME_FEAT_AE_ZDCN(v)		NVME_GET(v, FEAT_AE_ZDCN)
+#define NVME_FEAT_AE_PMDRLPCN(v)	NVME_GET(v, FEAT_AE_PMDRLPCN)
+#define NVME_FEAT_AE_ADLPCN(v)		NVME_GET(v, FEAT_AE_ADLPCN)
+#define NVME_FEAT_AE_HDLPCN(v)		NVME_GET(v, FEAT_AE_HDLPCN)
+#define NVME_FEAT_AE_DLPCN(v)		NVME_GET(v, FEAT_AE_DLPCN)
 
 static inline void nvme_feature_decode_async_event_config(__u32 value,
 			  __u8 *smart, bool *nan, bool *fw, bool *telem,
@@ -438,6 +448,15 @@ static inline void nvme_feature_decode_endurance_group_event_config(__u32 value,
 	*endgcw	= NVME_FEAT_EG_EGCW(value);
 }
 
+#define NVME_FEAT_PERFC_ATTRI(v) NVME_GET(v, FEAT_PERFC_ATTRI)
+#define NVME_FEAT_PERFC_RVSPA(v) NVME_GET(v, FEAT_PERFC_RVSPA)
+
+static inline void nvme_feature_decode_perf_characteristics(__u32 value, __u8 *attri, bool *rvspa)
+{
+	*attri = NVME_FEAT_PERFC_ATTRI(value);
+	*rvspa = NVME_FEAT_PERFC_RVSPA(value);
+}
+
 #define NVME_FEAT_SPM_PBSLC(v)		NVME_GET(v, FEAT_SPM_PBSLC)
 
 static inline void nvme_feature_decode_software_progress_marker(__u32 value,
@@ -560,8 +579,6 @@ char *kv_keymatch(const char *kv, const char *key);
  */
 char *startswith(const char *s, const char *prefix);
 
-#define min(x, y) ((x) > (y) ? (y) : (x))
-
 #define __round_mask(val, mult) ((__typeof__(val))((mult)-1))
 
 /**
@@ -638,9 +655,6 @@ enum nvme_version {
  */
 const char *nvme_get_version(enum nvme_version type);
 
-#define NVME_UUID_LEN_STRING	37  /* 1b4e28ba-2fa1-11d2-883f-0016d3cca427 + \0 */
-#define NVME_UUID_LEN		16
-
 /**
  * nvme_uuid_to_string - Return string represenation of encoded UUID
  * @uuid:	Binary encoded input UUID
